summaryrefslogtreecommitdiff
diff options
context:
space:
mode:
-rwxr-xr-x.gitignore1
-rw-r--r--node/Network.cpp101
-rw-r--r--node/Network.hpp14
-rw-r--r--node/Node.cpp2
-rw-r--r--node/Peer.cpp9
-rw-r--r--node/Peer.hpp2
6 files changed, 88 insertions, 41 deletions
diff --git a/.gitignore b/.gitignore
index 78a7ae78..2129d842 100755
--- a/.gitignore
+++ b/.gitignore
@@ -33,6 +33,7 @@
*.autosave
/ZeroTier One.dmg
/root-topology/*.secret
+/testnet/local-testnet/n????
/testnet/local-testnet/*/peers.persist
/testnet/local-testnet/*/authtoken.secret
/testnet/local-testnet/*/*.log
diff --git a/node/Network.cpp b/node/Network.cpp
index 74ae39ab..6ad6fcf8 100644
--- a/node/Network.cpp
+++ b/node/Network.cpp
@@ -111,33 +111,84 @@ SharedPtr<Network> Network::newInstance(const RuntimeEnvironment *renv,NodeConfi
return nw;
}
-bool Network::updateMulticastGroups()
+// Function object used by rescanMulticastGroups()
+class AnnounceMulticastGroupsToPeersWithActiveDirectPaths
{
- Mutex::Lock _l(_lock);
- EthernetTap *t = _tap;
- if (t) {
- // Grab current groups from the local tap
- bool updated = t->updateMulticastGroups(_myMulticastGroups);
-
- // Merge in learned groups from any hosts bridged in behind us
- for(std::map<MulticastGroup,uint64_t>::const_iterator mg(_multicastGroupsBehindMe.begin());mg!=_multicastGroupsBehindMe.end();++mg)
- _myMulticastGroups.insert(mg->first);
-
- // Add or remove BROADCAST group based on broadcast enabled netconf flag
- if ((_config)&&(_config->enableBroadcast())) {
- if (_myMulticastGroups.count(BROADCAST))
- return updated;
- else {
- _myMulticastGroups.insert(BROADCAST);
- return true;
+public:
+ AnnounceMulticastGroupsToPeersWithActiveDirectPaths(const RuntimeEnvironment *renv,Network *nw) :
+ RR(renv),
+ _now(Utils::now()),
+ _network(nw)
+ {}
+
+ inline void operator()(Topology &t,const SharedPtr<Peer> &p)
+ {
+ if ( ( (p->hasActiveDirectPath(_now)) && (_network->isAllowed(p->address())) ) || (_network->controller() == p->address()) || (t.isSupernode(p->address())) ) {
+ Packet outp(p->address(),RR->identity.address(),Packet::VERB_MULTICAST_LIKE);
+
+ std::set<MulticastGroup> mgs(_network->multicastGroups());
+ for(std::set<MulticastGroup>::iterator mg(mgs.begin());mg!=mgs.end();++mg) {
+ if ((outp.size() + 18) > ZT_UDP_DEFAULT_PAYLOAD_MTU) {
+ outp.armor(p->key(),true);
+ p->send(RR,outp.data(),outp.size(),_now);
+ outp.reset(p->address(),RR->identity.address(),Packet::VERB_MULTICAST_LIKE);
+ }
+
+ // network ID, MAC, ADI
+ outp.append((uint64_t)_network->id());
+ mg->mac().appendTo(outp);
+ outp.append((uint32_t)mg->adi());
+ }
+
+ if (outp.size() > ZT_PROTO_MIN_PACKET_LENGTH) {
+ outp.armor(p->key(),true);
+ p->send(RR,outp.data(),outp.size(),_now);
}
- } else {
- if (_myMulticastGroups.count(BROADCAST)) {
- _myMulticastGroups.erase(BROADCAST);
- return true;
- } else return updated;
}
- } else return false;
+ }
+
+private:
+ const RuntimeEnvironment *RR;
+ uint64_t _now;
+ Network *_network;
+};
+
+bool Network::rescanMulticastGroups()
+{
+ bool updated = false;
+
+ {
+ Mutex::Lock _l(_lock);
+ EthernetTap *t = _tap;
+ if (t) {
+ // Grab current groups from the local tap
+ updated = t->updateMulticastGroups(_myMulticastGroups);
+
+ // Merge in learned groups from any hosts bridged in behind us
+ for(std::map<MulticastGroup,uint64_t>::const_iterator mg(_multicastGroupsBehindMe.begin());mg!=_multicastGroupsBehindMe.end();++mg)
+ _myMulticastGroups.insert(mg->first);
+
+ // Add or remove BROADCAST group based on broadcast enabled netconf flag
+ if ((_config)&&(_config->enableBroadcast())) {
+ if (!_myMulticastGroups.count(BROADCAST)) {
+ _myMulticastGroups.insert(BROADCAST);
+ updated = true;
+ }
+ } else {
+ if (_myMulticastGroups.count(BROADCAST)) {
+ _myMulticastGroups.erase(BROADCAST);
+ updated = true;
+ }
+ }
+ }
+ }
+
+ if (updated) {
+ AnnounceMulticastGroupsToPeersWithActiveDirectPaths afunc(RR,this);
+ RR->topology->eachPeer<AnnounceMulticastGroupsToPeersWithActiveDirectPaths &>(afunc);
+ }
+
+ return updated;
}
bool Network::applyConfiguration(const SharedPtr<NetworkConfig> &conf)
@@ -449,6 +500,8 @@ void Network::threadMain()
t->setEnabled(_enabled);
}
}
+
+ rescanMulticastGroups();
}
void Network::_CBhandleTapData(void *arg,const MAC &from,const MAC &to,unsigned int etherType,const Buffer<4096> &data)
diff --git a/node/Node.cpp b/node/Node.cpp
index 9f95eaa7..8ca68a06 100644
--- a/node/Node.cpp
+++ b/node/Node.cpp
@@ -585,7 +585,7 @@ Node::ReasonForTermination Node::run()
try {
std::vector< SharedPtr<Network> > networks(RR->nc->networks());
for(std::vector< SharedPtr<Network> >::const_iterator nw(networks.begin());nw!=networks.end();++nw)
- (*nw)->updateMulticastGroups();
+ (*nw)->rescanMulticastGroups();
} catch (std::exception &exc) {
LOG("unexpected exception announcing multicast groups: %s",exc.what());
} catch ( ... ) {
diff --git a/node/Peer.cpp b/node/Peer.cpp
index 5f27da3f..8012c5ff 100644
--- a/node/Peer.cpp
+++ b/node/Peer.cpp
@@ -113,10 +113,11 @@ void Peer::receive(
}
}
- // Announce multicast groups of interest to direct peers if they are
- // considered authorized members of a given network. Also announce to
- // supernodes and network controllers. TODO: the former may go
- // obsolete with time as network controllers take over this role.
+ /* Announce multicast groups of interest to direct peers if they are
+ * considered authorized members of a given network. Also announce to
+ * supernodes and network controllers. The other place this is done
+ * is in rescanMulticastGroups() in Network, but that only sends something
+ * if a network's multicast groups change. */
if ((now - _lastAnnouncedTo) >= ((ZT_MULTICAST_LIKE_EXPIRE / 2) - 1000)) {
_lastAnnouncedTo = now;
diff --git a/node/Peer.hpp b/node/Peer.hpp
index 5d9395ee..deb15ac9 100644
--- a/node/Peer.hpp
+++ b/node/Peer.hpp
@@ -459,6 +459,8 @@ public:
}
private:
+ void _announceMulticastGroups(const RuntimeEnvironment *RR,uint64_t now);
+
unsigned char _key[ZT_PEER_SECRET_KEY_LENGTH];
Identity _id;
diff --git a/node/Network.hpp b/node/Network.hpp
index 7c21345e..90ac34d6 100644
--- a/node/Network.hpp
+++ b/node/Network.hpp
@@ -148,11 +148,11 @@ public:
}
/**
* Update multicast groups for this network's tap and announce changes
*
* @return True if internal multicast group set has changed since last update
*/
- bool updateMulticastGroups();
+ bool rescanMulticastGroups();
/**
* @return Latest set of multicast groups for this network's tap
@@ -164,16 +164,6 @@ public:
}
/**
- * @param mg Multicast group
- * @return True if this group is among those to which I am subscribed
- */
- inline bool wantMulticastGroup(const MulticastGroup &mg) const
- {
- Mutex::Lock _l(_lock);
- return (_myMulticastGroups.count(mg) > 0);
- }
-
- /**
* Apply a NetworkConfig to this network
*
* @param conf Configuration in NetworkConfig form
@@ -148,11 +148,11 @@ public:
}
/**
- * Update multicast groups for this network's tap and announce changes
+ * Rescan multicast groups for this network's tap and update peers on change
*
* @return True if internal multicast group set has changed since last update